from abc import abstractmethod


class SimpleSmoothing:

    @abstractmethod
    def setProbabilities(self,
                         nGram,
                         level: int):
        pass

    def setProbabilitiesGeneral(self, nGram):
        """
        Calculates the N-Gram probabilities with simple smoothing.
        :param nGram: N-Gram for which simple smoothing calculation is done.
        """
        self.setProbabilities(nGram, nGram.getN())



from NGram.NGram import NGram
from NGram.SimpleSmoothing import SimpleSmoothing


class NoSmoothing(SimpleSmoothing):

    def setProbabilities(self,
                         nGram: NGram,
                         level: int):
        """
        Calculates the N-Gram probabilities with no smoothing
        :param nGram: N-Gram for which no smoothing is done.
        :param level: Height of the NGram node.
        """
        nGram.setProbabilityWithPseudoCount(0.0, level)



class MultipleFile:

    __index: int
    __file_name_list: list
    __lines: list
    __line_index: int

    def __init__(self, fileList: list):
        """
        Constructor for MultipleFile class. Initializes the buffer reader with the first input file
        from the fileNameList. MultipleFile supports simple multipart file system, where a text file is divided
        into multiple files.
        :param fileList: A list of files.
        """
        self.__file_name_list = fileList
        file = open(self.__file_name_list[0])
        self.__lines = file.readlines()
        file.close()
        self.__line_index = 0
        self.__index = 0

    def readLine(self) -> str:
        """
        Reads a single line from the current file. If the end of file is reached for the current file,
        next file is opened and a single line from that file is read. If all files are read, the method
        returns null.
        :return: Read line from the current file.
        """
        if self.__line_index == len(self.__lines):
            self.__index = self.__index + 1
            file = open(self.__file_name_list[self.__index])
            self.__lines = file.readlines()
            file.close()
            self.__line_index = 0
        line = self.__lines[self.__line_index]
        self.__line_index = self.__line_index + 1
        return line



from Sampling.KFoldCrossValidation import KFoldCrossValidation

from NGram.NGram import NGram
from NGram.TrainedSmoothing import TrainedSmoothing
import math


class AdditiveSmoothing(TrainedSmoothing):

    __delta: float

    def __learnBestDelta(self,
                         nGrams: list,
                         kFoldCrossValidation: KFoldCrossValidation,
                         lowerBound: float) -> float:
        """
        The algorithm tries to optimize the best delta for a given corpus. The algorithm uses perplexity on the
        validation set as the optimization criterion.

        PARAMETERS
        ----------
        nGrams : list
            10 N-Grams learned for different folds of the corpus. nGrams[i] is the N-Gram trained with i'th train fold
            of the corpus.
        kFoldCrossValidation: KFoldCrossValidation
            Cross-validation data used in training and testing the N-grams.
        lowerBound : float
            Initial lower bound for optimizing the best delta.

        RETURNS
        -------
        float
            Best delta optimized with k-fold crossvalidation.
        """
        best_previous = -1
        upper_bound = 1
        best_delta = (lowerBound + upper_bound) / 2
        number_of_parts = 5
        while True:
            best_perplexity = 100000000
            value = lowerBound
            while value <= upper_bound:
                perplexity = 0
                for i in range(0, 10):
                    nGrams[i].setProbabilityWithPseudoCount(value, nGrams[i].getN())
                    perplexity += nGrams[i].getPerplexity(kFoldCrossValidation.getTestFold(i))
                if perplexity < best_perplexity:
                    best_perplexity = perplexity
                    best_delta = value
                value += (upper_bound - lowerBound) / number_of_parts
            lowerBound = self.newLowerBound(best_delta, lowerBound, upper_bound, number_of_parts)
            upper_bound = self.newUpperBound(best_delta, lowerBound, upper_bound, number_of_parts)
            if best_previous != -1:
                if math.fabs(best_previous - best_perplexity) / best_perplexity < 0.001:
                    break
            best_previous = best_perplexity
        return best_delta

    def learnParameters(self,
                        corpus: list,
                        N: int):
        """
        Wrapper function to learn the parameter (delta) in additive smoothing. The function first creates K NGrams
        with the train folds of the corpus. Then optimizes delta with respect to the test folds of the corpus.

        PARAMETERS
        ----------
        corpus : list
            Train corpus used to optimize delta parameter
        N : int
            N in N-Gram.
        """
        K = 10
        n_grams = []
        k_fold_cross_validation = KFoldCrossValidation(corpus, K, 0)
        for i in range(K):
            n_grams.append(NGram(N, k_fold_cross_validation.getTrainFold(i)))
        self.__delta = self.__learnBestDelta(n_grams, k_fold_cross_validation, 0.1)

    def setProbabilities(self,
                         nGram: NGram,
                         level: int):
        """
        Wrapper function to set the N-gram probabilities with additive smoothing.

        PARAMETERS
        ----------
        nGram : NGram
            N-Gram for which the probabilities will be set.
        level : int
            Level for which N-Gram probabilities will be set. Probabilities for different levels of the N-gram can be
            set with this function. If level = 1, N-Gram is treated as UniGram, if level = 2, N-Gram is treated as
            Bigram, etc.
        """
        nGram.setProbabilityWithPseudoCount(self.__delta, level)

    def getDelta(self) -> float:
        """
        Gets the best delta.

        RETURNS
        -------
        float
            learned best delta
        """
        return self.__delta
